# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.feedback_score_names import FeedbackScoreNames
from ..types.project_detailed import ProjectDetailed
from ..types.project_metric_response_public import ProjectMetricResponsePublic
from ..types.project_page_public import ProjectPagePublic
from ..types.project_public import ProjectPublic
from ..types.project_stats_summary import ProjectStatsSummary
from .raw_client import AsyncRawProjectsClient, RawProjectsClient
from .types.project_metric_request_public_interval import ProjectMetricRequestPublicInterval
from .types.project_metric_request_public_metric_type import ProjectMetricRequestPublicMetricType
from .types.project_update_visibility import ProjectUpdateVisibility
from .types.project_write_visibility import ProjectWriteVisibility

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ProjectsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawProjectsClient
        """
        return self._raw_client

    def find_projects(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectPagePublic:
        """
        Find projects

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPagePublic
            Project resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_projects()
        """
        _response = self._raw_client.find_projects(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    def create_project(
        self,
        *,
        name: str,
        visibility: typing.Optional[ProjectWriteVisibility] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create project

        Parameters
        ----------
        name : str

        visibility : typing.Optional[ProjectWriteVisibility]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.create_project(name='name', )
        """
        _response = self._raw_client.create_project(
            name=name, visibility=visibility, description=description, request_options=request_options
        )
        return _response.data

    def get_project_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> ProjectPublic:
        """
        Get project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPublic
            Project resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_project_by_id(id='id', )
        """
        _response = self._raw_client.get_project_by_id(id, request_options=request_options)
        return _response.data

    def delete_project_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.delete_project_by_id(id='id', )
        """
        _response = self._raw_client.delete_project_by_id(id, request_options=request_options)
        return _response.data

    def update_project(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        visibility: typing.Optional[ProjectUpdateVisibility] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update project by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        visibility : typing.Optional[ProjectUpdateVisibility]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.update_project(id='id', )
        """
        _response = self._raw_client.update_project(
            id, name=name, description=description, visibility=visibility, request_options=request_options
        )
        return _response.data

    def delete_projects_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete projects batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.delete_projects_batch(ids=['ids'], )
        """
        _response = self._raw_client.delete_projects_batch(ids=ids, request_options=request_options)
        return _response.data

    def find_feedback_score_names_by_project_ids(
        self, *, project_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> FeedbackScoreNames:
        """
        Find Feedback Score names By Project Ids

        Parameters
        ----------
        project_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackScoreNames
            Feedback Scores resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.find_feedback_score_names_by_project_ids()
        """
        _response = self._raw_client.find_feedback_score_names_by_project_ids(
            project_ids=project_ids, request_options=request_options
        )
        return _response.data

    def get_project_metrics(
        self,
        id: str,
        *,
        metric_type: typing.Optional[ProjectMetricRequestPublicMetricType] = OMIT,
        interval: typing.Optional[ProjectMetricRequestPublicInterval] = OMIT,
        interval_start: typing.Optional[dt.datetime] = OMIT,
        interval_end: typing.Optional[dt.datetime] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectMetricResponsePublic:
        """
        Gets specified metrics for a project

        Parameters
        ----------
        id : str

        metric_type : typing.Optional[ProjectMetricRequestPublicMetricType]

        interval : typing.Optional[ProjectMetricRequestPublicInterval]

        interval_start : typing.Optional[dt.datetime]

        interval_end : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectMetricResponsePublic
            Project Metrics

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_project_metrics(id='id', )
        """
        _response = self._raw_client.get_project_metrics(
            id,
            metric_type=metric_type,
            interval=interval,
            interval_start=interval_start,
            interval_end=interval_end,
            request_options=request_options,
        )
        return _response.data

    def get_project_stats(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectStatsSummary:
        """
        Get Project Stats

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectStatsSummary
            Project Stats

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.get_project_stats()
        """
        _response = self._raw_client.get_project_stats(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    def retrieve_project(
        self, *, name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectDetailed:
        """
        Retrieve project

        Parameters
        ----------
        name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectDetailed
            Project resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.projects.retrieve_project(name='name', )
        """
        _response = self._raw_client.retrieve_project(name=name, request_options=request_options)
        return _response.data


class AsyncProjectsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawProjectsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawProjectsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawProjectsClient
        """
        return self._raw_client

    async def find_projects(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectPagePublic:
        """
        Find projects

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPagePublic
            Project resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_projects()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_projects(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    async def create_project(
        self,
        *,
        name: str,
        visibility: typing.Optional[ProjectWriteVisibility] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create project

        Parameters
        ----------
        name : str

        visibility : typing.Optional[ProjectWriteVisibility]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.create_project(name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_project(
            name=name, visibility=visibility, description=description, request_options=request_options
        )
        return _response.data

    async def get_project_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectPublic:
        """
        Get project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectPublic
            Project resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_project_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_project_by_id(id, request_options=request_options)
        return _response.data

    async def delete_project_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.delete_project_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_project_by_id(id, request_options=request_options)
        return _response.data

    async def update_project(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        visibility: typing.Optional[ProjectUpdateVisibility] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update project by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        visibility : typing.Optional[ProjectUpdateVisibility]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.update_project(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_project(
            id, name=name, description=description, visibility=visibility, request_options=request_options
        )
        return _response.data

    async def delete_projects_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete projects batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.delete_projects_batch(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_projects_batch(ids=ids, request_options=request_options)
        return _response.data

    async def find_feedback_score_names_by_project_ids(
        self, *, project_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> FeedbackScoreNames:
        """
        Find Feedback Score names By Project Ids

        Parameters
        ----------
        project_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        FeedbackScoreNames
            Feedback Scores resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.find_feedback_score_names_by_project_ids()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_feedback_score_names_by_project_ids(
            project_ids=project_ids, request_options=request_options
        )
        return _response.data

    async def get_project_metrics(
        self,
        id: str,
        *,
        metric_type: typing.Optional[ProjectMetricRequestPublicMetricType] = OMIT,
        interval: typing.Optional[ProjectMetricRequestPublicInterval] = OMIT,
        interval_start: typing.Optional[dt.datetime] = OMIT,
        interval_end: typing.Optional[dt.datetime] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectMetricResponsePublic:
        """
        Gets specified metrics for a project

        Parameters
        ----------
        id : str

        metric_type : typing.Optional[ProjectMetricRequestPublicMetricType]

        interval : typing.Optional[ProjectMetricRequestPublicInterval]

        interval_start : typing.Optional[dt.datetime]

        interval_end : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectMetricResponsePublic
            Project Metrics

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_project_metrics(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_project_metrics(
            id,
            metric_type=metric_type,
            interval=interval,
            interval_start=interval_start,
            interval_end=interval_end,
            request_options=request_options,
        )
        return _response.data

    async def get_project_stats(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ProjectStatsSummary:
        """
        Get Project Stats

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectStatsSummary
            Project Stats

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.get_project_stats()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_project_stats(
            page=page, size=size, name=name, sorting=sorting, request_options=request_options
        )
        return _response.data

    async def retrieve_project(
        self, *, name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> ProjectDetailed:
        """
        Retrieve project

        Parameters
        ----------
        name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ProjectDetailed
            Project resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.projects.retrieve_project(name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.retrieve_project(name=name, request_options=request_options)
        return _response.data



import logging
from typing import Any, Dict, List, Optional

import pydantic
from openai.types.chat import chat_completion_chunk

from opik import logging_messages

LOGGER = logging.getLogger(__name__)


class ChatCompletionChunksAggregated(pydantic.BaseModel):
    choices: List[Dict[str, Any]]
    created: int
    id: str
    model: str
    object: str
    system_fingerprint: Optional[str]
    usage: Optional[Dict[str, Any]]


def aggregate(
    items: List[chat_completion_chunk.ChatCompletionChunk],
) -> Optional[ChatCompletionChunksAggregated]:
    # TODO: check if there are scenarios when stream contains more than one choice
    try:
        first_chunk = items[0]

        aggregated_response = {
            "choices": [{"index": 0, "message": {"role": "", "content": ""}}],
            "created": first_chunk.created,
            "id": first_chunk.id,
            "model": first_chunk.model,
            "object": "chat.completion",
            "system_fingerprint": first_chunk.system_fingerprint,
            "usage": None,
        }

        text_chunks: List[str] = []

        for chunk in items:
            if chunk.choices and chunk.choices[0].delta:
                delta = chunk.choices[0].delta

                if (
                    delta.role
                    and not aggregated_response["choices"][0]["message"]["role"]
                ):
                    aggregated_response["choices"][0]["message"]["role"] = delta.role

                if delta.content:
                    text_chunks.append(delta.content)

            if chunk.choices and chunk.choices[0].finish_reason:
                aggregated_response["choices"][0]["finish_reason"] = chunk.choices[
                    0
                ].finish_reason

            if chunk.usage:
                aggregated_response["usage"] = chunk.usage.model_dump()

        aggregated_response["choices"][0]["message"]["content"] = "".join(text_chunks)
        result = ChatCompletionChunksAggregated(**aggregated_response)

        return result
    except Exception as exception:
        LOGGER.error(
            logging_messages.FAILED_TO_PARSE_OPENAI_STREAM_CONTENT,
            str(exception),
            exc_info=True,
        )
        return None



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from .raw_client import AsyncRawRedirectClient, RawRedirectClient


class RedirectClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawRedirectClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawRedirectClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawRedirectClient
        """
        return self._raw_client

    def datasets_redirect(
        self,
        *,
        dataset_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create dataset redirect url

        Parameters
        ----------
        dataset_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.redirect.datasets_redirect(dataset_id='dataset_id', path='path', )
        """
        _response = self._raw_client.datasets_redirect(
            dataset_id=dataset_id, path=path, workspace_name=workspace_name, request_options=request_options
        )
        return _response.data

    def experiments_redirect(
        self,
        *,
        dataset_id: str,
        experiment_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create experiment redirect url

        Parameters
        ----------
        dataset_id : str

        experiment_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.redirect.experiments_redirect(dataset_id='dataset_id', experiment_id='experiment_id', path='path', )
        """
        _response = self._raw_client.experiments_redirect(
            dataset_id=dataset_id,
            experiment_id=experiment_id,
            path=path,
            workspace_name=workspace_name,
            request_options=request_options,
        )
        return _response.data

    def projects_redirect(
        self,
        *,
        trace_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create project redirect url

        Parameters
        ----------
        trace_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.redirect.projects_redirect(trace_id='trace_id', path='path', )
        """
        _response = self._raw_client.projects_redirect(
            trace_id=trace_id, path=path, workspace_name=workspace_name, request_options=request_options
        )
        return _response.data


class AsyncRedirectClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawRedirectClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawRedirectClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawRedirectClient
        """
        return self._raw_client

    async def datasets_redirect(
        self,
        *,
        dataset_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create dataset redirect url

        Parameters
        ----------
        dataset_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.redirect.datasets_redirect(dataset_id='dataset_id', path='path', )
        asyncio.run(main())
        """
        _response = await self._raw_client.datasets_redirect(
            dataset_id=dataset_id, path=path, workspace_name=workspace_name, request_options=request_options
        )
        return _response.data

    async def experiments_redirect(
        self,
        *,
        dataset_id: str,
        experiment_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create experiment redirect url

        Parameters
        ----------
        dataset_id : str

        experiment_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.redirect.experiments_redirect(dataset_id='dataset_id', experiment_id='experiment_id', path='path', )
        asyncio.run(main())
        """
        _response = await self._raw_client.experiments_redirect(
            dataset_id=dataset_id,
            experiment_id=experiment_id,
            path=path,
            workspace_name=workspace_name,
            request_options=request_options,
        )
        return _response.data

    async def projects_redirect(
        self,
        *,
        trace_id: str,
        path: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create project redirect url

        Parameters
        ----------
        trace_id : str

        path : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.redirect.projects_redirect(trace_id='trace_id', path='path', )
        asyncio.run(main())
        """
        _response = await self._raw_client.projects_redirect(
            trace_id=trace_id, path=path, workspace_name=workspace_name, request_options=request_options
        )
        return _response.data



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .complete_multipart_upload_request_entity_type import CompleteMultipartUploadRequestEntityType
from .multipart_upload_part import MultipartUploadPart


class CompleteMultipartUploadRequest(UniversalBaseModel):
    file_name: str
    project_name: typing.Optional[str] = pydantic.Field(default=None)
    """
    If null, the default project is used
    """

    entity_type: CompleteMultipartUploadRequestEntityType
    entity_id: str
    container_id: typing.Optional[str] = None
    file_size: int
    mime_type: typing.Optional[str] = None
    upload_id: str
    uploaded_file_parts: typing.List[MultipartUploadPart]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import dataclasses
import os
from typing import Optional

from ..message_processing import messages
from ..types import AttachmentEntityType


@dataclasses.dataclass
class FileUploadOptions:
    file_path: str
    file_name: str
    file_size: int
    mime_type: Optional[str]
    entity_type: AttachmentEntityType
    entity_id: str
    project_name: str
    encoded_url_override: str


def file_upload_options_from_attachment(
    attachment: messages.CreateAttachmentMessage,
) -> FileUploadOptions:
    file_size = os.path.getsize(attachment.file_path)

    return FileUploadOptions(
        file_path=attachment.file_path,
        file_name=attachment.file_name,
        file_size=file_size,
        mime_type=attachment.mime_type,
        entity_type=attachment.entity_type,
        entity_id=attachment.entity_id,
        project_name=attachment.project_name,
        encoded_url_override=attachment.encoded_url_override,
    )



from typing import Union, cast

from . import base_batcher
from .. import messages


class CreateSpanMessageBatcher(base_batcher.BaseBatcher):
    def _create_batch_from_accumulated_messages(
        self,
    ) -> messages.CreateSpansBatchMessage:
        return messages.CreateSpansBatchMessage(batch=self._accumulated_messages)  # type: ignore

    def add(self, message: messages.CreateSpanMessage) -> None:  # type: ignore
        return super().add(message)


class CreateTraceMessageBatcher(base_batcher.BaseBatcher):
    def _create_batch_from_accumulated_messages(
        self,
    ) -> messages.CreateTraceBatchMessage:
        return messages.CreateTraceBatchMessage(batch=self._accumulated_messages)  # type: ignore

    def add(self, message: messages.CreateTraceMessage) -> None:  # type: ignore
        return super().add(message)


class BaseAddFeedbackScoresBatchMessageBatcher(base_batcher.BaseBatcher):
    def _create_batch_from_accumulated_messages(  # type: ignore
        self,
    ) -> Union[
        messages.AddSpanFeedbackScoresBatchMessage,
        messages.AddTraceFeedbackScoresBatchMessage,
    ]:
        return super()._create_batch_from_accumulated_messages()  # type: ignore

    def add(  # type: ignore
        self,
        message: Union[
            messages.AddSpanFeedbackScoresBatchMessage,
            messages.AddTraceFeedbackScoresBatchMessage,
        ],
    ) -> None:
        with self._lock:
            new_messages = message.batch
            n_new_messages = len(new_messages)
            n_accumulated_messages = len(self._accumulated_messages)

            if n_new_messages + n_accumulated_messages >= self._max_batch_size:
                free_space_in_accumulator = (
                    self._max_batch_size - n_accumulated_messages
                )

                messages_that_fit_in_batch = new_messages[:free_space_in_accumulator]
                messages_that_dont_fit_in_batch = new_messages[
                    free_space_in_accumulator:
                ]

                self._accumulated_messages += messages_that_fit_in_batch
                new_messages = messages_that_dont_fit_in_batch
                self.flush()

            self._accumulated_messages += new_messages


class AddSpanFeedbackScoresBatchMessageBatcher(
    BaseAddFeedbackScoresBatchMessageBatcher
):
    def _create_batch_from_accumulated_messages(
        self,
    ) -> messages.AddSpanFeedbackScoresBatchMessage:  # type: ignore
        return messages.AddSpanFeedbackScoresBatchMessage(
            batch=self._accumulated_messages,  # type: ignore
            supports_batching=False,
        )


class AddTraceFeedbackScoresBatchMessageBatcher(
    BaseAddFeedbackScoresBatchMessageBatcher
):
    def _create_batch_from_accumulated_messages(
        self,
    ) -> messages.AddTraceFeedbackScoresBatchMessage:  # type: ignore
        return messages.AddTraceFeedbackScoresBatchMessage(
            batch=self._accumulated_messages,  # type: ignore
            supports_batching=False,
        )


class GuardrailBatchMessageBatcher(base_batcher.BaseBatcher):
    def _create_batch_from_accumulated_messages(
        self,
    ) -> messages.GuardrailBatchMessage:  # type: ignore
        batch = []

        for batch_message in self._accumulated_messages:
            batch_of_message_item = cast(
                messages.GuardrailBatchMessage, batch_message
            ).batch
            batch.extend(batch_of_message_item)

        return messages.GuardrailBatchMessage(batch=batch, supports_batching=False)  # type: ignore

    def add(self, message: messages.GuardrailBatchMessage) -> None:  # type: ignore
        return super().add(message)



import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from typing_extensions import override

import openai

from opik import LLMProvider, dict_utils, llm_usage
from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator
from openai.types import responses as openai_responses

from . import stream_patchers

LOGGER = logging.getLogger(__name__)

KWARGS_KEYS_TO_LOG_AS_INPUTS = [
    "input",
    "instructions",
    "reasoning",
    "previous_response_id",
]
RESPONSE_KEYS_TO_LOG_AS_OUTPUT = [
    "output",
    "choices",
    "reasoning",
]


class OpenaiResponsesTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    An implementation of BaseTrackDecorator designed specifically for tracking
    calls of OpenAI's `responses.create` and `responses.parse` functions.
    """

    def __init__(self) -> None:
        super().__init__()
        self.provider = "openai"

    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Optional[Tuple],
        kwargs: Optional[Dict[str, Any]],
    ) -> arguments_helpers.StartSpanParameters:
        assert (
            kwargs is not None
        ), "Expected kwargs to be not None in responses.create(**kwargs) or responses.parse(**kwargs)"

        name = track_options.name if track_options.name is not None else func.__name__

        metadata = track_options.metadata if track_options.metadata is not None else {}

        input, new_metadata = dict_utils.split_dict_by_keys(
            kwargs, keys=KWARGS_KEYS_TO_LOG_AS_INPUTS
        )
        metadata = dict_utils.deepmerge(metadata, new_metadata)
        metadata.update(
            {
                "created_from": "openai",
                "type": "openai_responses",
            }
        )

        tags = ["openai"]
        model = kwargs.get("model", None)
        result = arguments_helpers.StartSpanParameters(
            name=name,
            input=input,
            type=track_options.type,
            tags=tags,
            metadata=metadata,
            project_name=track_options.project_name,
            model=model,
            provider=self.provider,
        )

        return result

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        assert isinstance(output, openai_responses.Response)

        result_dict = output.model_dump(mode="json")

        output_data, metadata = dict_utils.split_dict_by_keys(
            result_dict, RESPONSE_KEYS_TO_LOG_AS_OUTPUT
        )

        opik_usage = None
        if result_dict.get("usage") is not None:
            opik_usage = llm_usage.try_build_opik_usage_or_log_error(
                provider=LLMProvider.OPENAI,
                usage=result_dict["usage"],
                logger=LOGGER,
                error_message="Failed to log token usage from openai responses call",
            )

        model = result_dict.get("model", None)

        result = arguments_helpers.EndSpanParameters(
            output=output_data,
            usage=opik_usage,
            metadata=metadata,
            model=model,
            provider=self.provider,
        )

        return result

    @override
    def _streams_handler(  # type: ignore
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[
            Callable[
                [List[openai_responses.ResponseStreamEvent]], openai_responses.Response
            ]
        ],
    ) -> Union[
        None,
        openai.Stream,
        openai.AsyncStream,
    ]:
        assert (
            generations_aggregator is not None
        ), "OpenAI decorator will always get aggregator function as input"

        if isinstance(output, openai.Stream):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            return stream_patchers.patch_sync_stream(
                stream=output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        if isinstance(output, openai.AsyncStream):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            return stream_patchers.patch_async_stream(
                stream=output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        NOT_A_STREAM = None
        return NOT_A_STREAM



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .json_node_detail import JsonNodeDetail
from .prompt_version_detail_type import PromptVersionDetailType


class PromptVersionDetail(UniversalBaseModel):
    id: typing.Optional[str] = pydantic.Field(default=None)
    """
    version unique identifier, generated if absent
    """

    prompt_id: typing.Optional[str] = None
    commit: typing.Optional[str] = pydantic.Field(default=None)
    """
    version short unique identifier, generated if absent. it must be 8 characters long
    """

    template: str
    metadata: typing.Optional[JsonNodeDetail] = None
    type: typing.Optional[PromptVersionDetailType] = None
    change_description: typing.Optional[str] = None
    variables: typing.Optional[typing.List[str]] = None
    created_at: typing.Optional[dt.datetime] = None
    created_by: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from typing import List, Optional

import httpx

from . import queue_consumer, messages, message_processors, message_queue, streamer
from ..file_upload import upload_manager, base_upload_manager
from ..rest_api import client as rest_api_client
from .batching import batch_manager_constuctors


def construct_online_streamer(
    rest_client: rest_api_client.OpikApi,
    httpx_client: httpx.Client,
    use_batching: bool,
    file_upload_worker_count: int,
    n_consumers: int,
    max_queue_size: int,
) -> streamer.Streamer:
    message_processor = message_processors.MessageSender(rest_client=rest_client)

    file_uploader = upload_manager.FileUploadManager(
        rest_client=rest_client,
        httpx_client=httpx_client,
        worker_count=file_upload_worker_count,
    )

    return construct_streamer(
        message_processor=message_processor,
        file_upload_manager=file_uploader,
        n_consumers=n_consumers,
        use_batching=use_batching,
        max_queue_size=max_queue_size,
    )


def construct_streamer(
    message_processor: message_processors.BaseMessageProcessor,
    file_upload_manager: base_upload_manager.BaseFileUploadManager,
    n_consumers: int,
    use_batching: bool,
    max_queue_size: Optional[int],
) -> streamer.Streamer:
    message_queue_: message_queue.MessageQueue[messages.BaseMessage] = (
        message_queue.MessageQueue(max_length=max_queue_size)
    )

    queue_consumers: List[queue_consumer.QueueConsumer] = [
        queue_consumer.QueueConsumer(
            queue=message_queue_,
            message_processor=message_processor,
            name=f"QueueConsumerThread_{i}",
        )
        for i in range(n_consumers)
    ]

    batch_manager = (
        batch_manager_constuctors.create_batch_manager(message_queue_)
        if use_batching
        else None
    )

    streamer_ = streamer.Streamer(
        queue=message_queue_,
        queue_consumers=queue_consumers,
        batch_manager=batch_manager,
        file_upload_manager=file_upload_manager,
    )

    return streamer_



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.json_node_write import JsonNodeWrite
from ..types.optimization_page_public import OptimizationPagePublic
from ..types.optimization_public import OptimizationPublic
from .raw_client import AsyncRawOptimizationsClient, RawOptimizationsClient
from .types.optimization_update_status import OptimizationUpdateStatus
from .types.optimization_write_status import OptimizationWriteStatus

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class OptimizationsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawOptimizationsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawOptimizationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawOptimizationsClient
        """
        return self._raw_client

    def find_optimizations(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OptimizationPagePublic:
        """
        Find optimizations

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizationPagePublic
            Optimizations resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.optimizations.find_optimizations()
        """
        _response = self._raw_client.find_optimizations(
            page=page,
            size=size,
            dataset_id=dataset_id,
            name=name,
            dataset_deleted=dataset_deleted,
            request_options=request_options,
        )
        return _response.data

    def create_optimization(
        self,
        *,
        dataset_name: str,
        objective_name: str,
        status: OptimizationWriteStatus,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create optimization

        Parameters
        ----------
        dataset_name : str

        objective_name : str

        status : OptimizationWriteStatus

        id : typing.Optional[str]

        name : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.optimizations.create_optimization(dataset_name='dataset_name', objective_name='objective_name', status="running", )
        """
        _response = self._raw_client.create_optimization(
            dataset_name=dataset_name,
            objective_name=objective_name,
            status=status,
            id=id,
            name=name,
            metadata=metadata,
            request_options=request_options,
        )
        return _response.data

    def delete_optimizations_by_id(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete optimizations by id

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.optimizations.delete_optimizations_by_id(ids=['ids'], )
        """
        _response = self._raw_client.delete_optimizations_by_id(ids=ids, request_options=request_options)
        return _response.data

    def get_optimization_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> OptimizationPublic:
        """
        Get optimization by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizationPublic
            Optimization resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.optimizations.get_optimization_by_id(id='id', )
        """
        _response = self._raw_client.get_optimization_by_id(id, request_options=request_options)
        return _response.data

    def update_optimizations_by_id(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        status: typing.Optional[OptimizationUpdateStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update optimization by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        status : typing.Optional[OptimizationUpdateStatus]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.optimizations.update_optimizations_by_id(id='id', )
        """
        _response = self._raw_client.update_optimizations_by_id(
            id, name=name, status=status, request_options=request_options
        )
        return _response.data


class AsyncOptimizationsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawOptimizationsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawOptimizationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawOptimizationsClient
        """
        return self._raw_client

    async def find_optimizations(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OptimizationPagePublic:
        """
        Find optimizations

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizationPagePublic
            Optimizations resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.optimizations.find_optimizations()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_optimizations(
            page=page,
            size=size,
            dataset_id=dataset_id,
            name=name,
            dataset_deleted=dataset_deleted,
            request_options=request_options,
        )
        return _response.data

    async def create_optimization(
        self,
        *,
        dataset_name: str,
        objective_name: str,
        status: OptimizationWriteStatus,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create optimization

        Parameters
        ----------
        dataset_name : str

        objective_name : str

        status : OptimizationWriteStatus

        id : typing.Optional[str]

        name : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.optimizations.create_optimization(dataset_name='dataset_name', objective_name='objective_name', status="running", )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_optimization(
            dataset_name=dataset_name,
            objective_name=objective_name,
            status=status,
            id=id,
            name=name,
            metadata=metadata,
            request_options=request_options,
        )
        return _response.data

    async def delete_optimizations_by_id(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete optimizations by id

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.optimizations.delete_optimizations_by_id(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_optimizations_by_id(ids=ids, request_options=request_options)
        return _response.data

    async def get_optimization_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> OptimizationPublic:
        """
        Get optimization by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OptimizationPublic
            Optimization resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.optimizations.get_optimization_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_optimization_by_id(id, request_options=request_options)
        return _response.data

    async def update_optimizations_by_id(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        status: typing.Optional[OptimizationUpdateStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update optimization by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        status : typing.Optional[OptimizationUpdateStatus]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.optimizations.update_optimizations_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_optimizations_by_id(
            id, name=name, status=status, request_options=request_options
        )
        return _response.data



import logging
import uuid
from typing import Any, Dict, List, Optional, Union

from google.adk.agents.callback_context import CallbackContext
from google.adk.models import LlmRequest, LlmResponse
from google.adk.tools.base_tool import BaseTool
from google.adk.tools.tool_context import ToolContext
from opik import context_storage
from opik.api_objects import helpers, opik_client, trace, span
from opik.types import DistributedTraceHeadersDict, SpanType

from . import llm_response_wrapper
from . import helpers as adk_helpers

LOGGER = logging.getLogger(__name__)

SpanOrTraceData = Union[span.SpanData, trace.TraceData]


class OpikTracer:
    def __init__(
        self,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        project_name: Optional[str] = None,
        distributed_headers: Optional[DistributedTraceHeadersDict] = None,
    ):
        self.name = name
        self.tags = tags
        self.metadata = metadata or {}
        self.project_name = project_name
        self.distributed_headers = distributed_headers
        self._client = opik_client.get_client_cached()

        self._last_model_output: Optional[Dict[str, Any]] = None

        # Use OpikContextStorage instance instead of global context storage module
        # in case we need to use different context storage for ADK in the future
        self._context_storage = context_storage.get_current_context_instance()

        self._opik_client = opik_client.get_client_cached()

        # monkey patch LLMResponse to store usage_metadata
        old_function = LlmResponse.create
        create_wrapper = llm_response_wrapper.LlmResponseCreateWrapper(old_function)
        LlmResponse.create = create_wrapper

    def _attach_span_to_existing_span(
        self,
        current_span_data: span.SpanData,
        name: str,
        input: Dict[str, Any],
        type: SpanType,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        project_name = helpers.resolve_child_span_project_name(
            parent_project_name=current_span_data.project_name,
            child_project_name=self.project_name,
        )

        span_data = span.SpanData(
            trace_id=current_span_data.trace_id,
            parent_span_id=current_span_data.id,
            name=name,
            input=input,
            type=type,
            project_name=project_name,
            metadata=self.metadata
            if metadata is None
            else {**self.metadata, **metadata},
        )
        self._set_current_context_data(span_data)

    def _attach_span_to_existing_trace(
        self,
        current_trace_data: trace.TraceData,
        name: str,
        input: Dict[str, Any],
        type: SpanType,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        project_name = helpers.resolve_child_span_project_name(
            parent_project_name=current_trace_data.project_name,
            child_project_name=self.project_name,
        )
        span_data = span.SpanData(
            trace_id=current_trace_data.id,
            parent_span_id=None,
            name=name,
            input=input,
            type=type,
            project_name=project_name,
            metadata=self.metadata
            if metadata is None
            else {**self.metadata, **metadata},
        )
        self._set_current_context_data(span_data)

    def _start_trace(
        self,
        new_trace_data: trace.TraceData,
    ) -> None:
        self._set_current_context_data(new_trace_data)

    def _end_current_trace(self) -> None:
        if (trace_data := self._context_storage.get_trace_data()) is not None:
            trace_data.init_end_time()
            self._opik_client.trace(**trace_data.__dict__)

            self._context_storage.set_trace_data(None)

    def _end_current_span(
        self,
    ) -> None:
        if (span_data := self._context_storage.top_span_data()) is not None:
            span_data.init_end_time()
            self._opik_client.span(**span_data.__dict__)

            self._context_storage.pop_span_data()

    def _set_current_context_data(self, value: SpanOrTraceData) -> None:
        if isinstance(value, span.SpanData):
            self._context_storage.add_span_data(value)
        elif isinstance(value, trace.TraceData):
            self._context_storage.set_trace_data(value)
        else:
            raise ValueError(f"Invalid context type: {type(value)}")

    def before_agent_callback(
        self, callback_context: CallbackContext, *args: Any, **kwargs: Any
    ) -> None:
        if "opik_thread_id" in callback_context.state:
            thread_id = callback_context.state["opik_thread_id"]
        else:
            thread_id = str(uuid.uuid4())
            callback_context.state["opik_thread_id"] = thread_id

        trace_metadata = self.metadata.copy()
        trace_metadata["adk_invocation_id"] = callback_context.invocation_id

        user_input = adk_helpers.convert_adk_base_model_to_dict(
            callback_context.user_content
        )
        name = self.name or callback_context.agent_name

        if (current_span_data := self._context_storage.top_span_data()) is not None:
            self._attach_span_to_existing_span(
                current_span_data=current_span_data,
                name=name,
                input=user_input,
                type="general",
                metadata=self.metadata,
            )
        elif (current_trace_data := self._context_storage.get_trace_data()) is not None:
            self._attach_span_to_existing_trace(
                current_trace_data=current_trace_data,
                name=name,
                input=user_input,
                type="general",
                metadata=self.metadata,
            )
        else:
            new_trace_data = trace.TraceData(
                name=name,
                input=user_input,
                metadata=self.metadata,
                project_name=self.project_name,
                thread_id=thread_id,
            )
            self._start_trace(new_trace_data)

    def after_agent_callback(
        self, callback_context: CallbackContext, *args: Any, **kwargs: Any
    ) -> None:
        output = self._last_model_output

        if (span_data := self._context_storage.top_span_data()) is None:
            trace_data = self._context_storage.get_trace_data()
            assert trace_data is not None
            trace_data.update(output=output).init_end_time()
            self._end_current_trace()
            self._last_model_output = None
        else:
            span_data.update(output=output).init_end_time()
            self._end_current_span()

    def before_model_callback(
        self,
        callback_context: CallbackContext,
        llm_request: LlmRequest,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        input = adk_helpers.convert_adk_base_model_to_dict(llm_request)
        if (current_span_data := self._context_storage.top_span_data()) is not None:
            self._attach_span_to_existing_span(
                current_span_data=current_span_data,
                name=llm_request.model,
                input=input,
                type="llm",
                metadata=self.metadata,
            )
        else:
            current_trace_data = self._context_storage.get_trace_data()
            assert current_trace_data is not None
            self._attach_span_to_existing_trace(
                current_trace_data=current_trace_data,
                name=llm_request.model,
                input=input,
                type="llm",
                metadata=self.metadata,
            )

    def after_model_callback(
        self,
        callback_context: CallbackContext,
        llm_response: LlmResponse,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            # Ignore partial chunks, ADK will call this method with the full
            # response at the end
            if llm_response.partial is True:
                return
        except Exception:
            LOGGER.debug("Error checking for partial chunks", exc_info=True)

        output = adk_helpers.convert_adk_base_model_to_dict(llm_response)
        usage_data = llm_response_wrapper.pop_llm_usage_data(output)
        if usage_data is not None:
            model = usage_data.model
            provider = usage_data.provider
            usage = usage_data.opik_usage
        else:
            model = None
            provider = None
            usage = None

        self._last_model_output = output

        span_data = self._context_storage.top_span_data()
        assert span_data is not None

        span_data.update(
            output=output,
            usage=usage,
            model=model,
            provider=provider,
        )
        self._end_current_span()

    def before_tool_callback(
        self,
        tool: BaseTool,
        args: Dict[str, Any],
        tool_context: ToolContext,
        *other_args: Any,
        **kwargs: Any,
    ) -> None:
        metadata = {"function_call_id": tool_context.function_call_id}

        if (current_span_data := self._context_storage.top_span_data()) is not None:
            self._attach_span_to_existing_span(
                current_span_data=current_span_data,
                name=tool.name,
                input=args,
                type="tool",
                metadata=metadata,
            )
        else:
            current_trace_data = self._context_storage.get_trace_data()
            assert current_trace_data is not None
            self._attach_span_to_existing_trace(
                current_trace_data=current_trace_data,
                name=tool.name,
                input=args,
                type="tool",
                metadata=metadata,
            )

    def after_tool_callback(
        self,
        tool: BaseTool,
        args: Dict[str, Any],
        tool_context: ToolContext,
        tool_response: Any,
        *other_args: Any,
        **kwargs: Any,
    ) -> None:
        current_span_data = self._context_storage.top_span_data()
        assert current_span_data is not None
        if isinstance(tool_response, dict):
            current_span_data.update(output=tool_response)
        else:
            current_span_data.update(output={"output": tool_response})
        self._end_current_span()



import functools
import getpass
import logging
import os
import platform
import socket
import sys
from importlib import metadata
from typing import Dict, Literal
import tqdm
from tqdm.utils import Comparable

import opik.config
from opik import url_helpers

LOGGER = logging.getLogger(__name__)


def get_pid() -> int:
    return os.getpid()


@functools.lru_cache
def get_hostname() -> str:
    return socket.gethostname()


@functools.lru_cache
def get_user() -> str:
    try:
        return getpass.getuser()
    except Exception:
        LOGGER.debug(
            "Unknown exception getting the user from the system", exc_info=True
        )
        return "unknown"


@functools.lru_cache
def get_os() -> str:
    return platform.platform(aliased=True)


@functools.lru_cache
def get_os_type() -> str:
    return platform.system()


@functools.lru_cache
def get_python_version_verbose() -> str:
    return sys.version


@functools.lru_cache
def get_python_version() -> str:
    return platform.python_version()


def in_pytest() -> bool:
    return "pytest" in sys.modules


@functools.lru_cache
def in_github_actions() -> bool:
    return "GITHUB_ACTIONS" in os.environ


@functools.lru_cache
def in_aws_lambda() -> bool:
    return "LAMBDA_TASK_ROOT" in os.environ


def get_installation_type() -> Literal["cloud", "self-hosted", "local"]:
    config = opik.config.OpikConfig()
    url_override = config.url_override
    if url_helpers.get_base_url(url_override) == url_helpers.get_base_url(
        opik.config.OPIK_URL_CLOUD
    ):
        return "cloud"

    if "localhost" in url_override:
        return "local"

    return "self-hosted"


@functools.lru_cache
def in_jupyter() -> bool:
    """
    Check to see if code is running in a Jupyter environment,
    including jupyter notebook, lab, or console.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    if ipy is None or not hasattr(ipy, "kernel"):
        return False
    else:
        return True


@functools.lru_cache
def in_ipython() -> bool:
    """
    Check to see if code is running in an IPython environment.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    if ipy is None:
        return False
    else:
        return True


@functools.lru_cache
def in_colab() -> bool:
    """
    Check to see if code is running in Google colab.
    """
    try:
        import IPython
    except Exception:
        return False

    ipy = IPython.get_ipython()
    return "google.colab" in str(ipy)


@functools.lru_cache
def get_installed_packages() -> Dict[str, str]:
    """
    Retrieve a dictionary of installed packages with their versions.
    """
    installed_packages = {
        pkg.metadata["Name"]: pkg.version for pkg in metadata.distributions()
    }
    return installed_packages


def get_tqdm_for_current_environment() -> Comparable:
    """
    Get a tqdm progress bar for your environment.
    """
    if in_jupyter() or in_colab():
        return tqdm.tqdm_notebook
    else:
        return tqdm.tqdm



from typing import Any, Optional

from .. import base_metric, score_result


class Equals(base_metric.BaseMetric):
    """
    A metric that checks if an output string exactly matches an expected output string.

    This metric returns a score of 1.0 if the strings match exactly, and 0.0 otherwise.
    The comparison can be made case-sensitive or case-insensitive.

    Args:
        case_sensitive: Whether the comparison should be case-sensitive. Defaults to False.
        name: The name of the metric. Defaults to "equals_metric".
        track: Whether to track the metric. Defaults to True.
        project_name: Optional project name to track the metric in for the cases when there are no parent span/trace to inherit project name from.

    Example:
        >>> from opik.evaluation.metrics import Equals
        >>> equals_metric = Equals(case_sensitive=True)
        >>> result = equals_metric.score("Hello, World!", "Hello, World!")
        >>> print(result.value)
        1.0
        >>> result = equals_metric.score("Hello, World!", "hello, world!")
        >>> print(result.value)
        0.0
    """

    def __init__(
        self,
        case_sensitive: bool = False,
        name: str = "equals_metric",
        track: bool = True,
        project_name: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            track=track,
            project_name=project_name,
        )
        self._case_sensitive = case_sensitive

    def score(
        self, output: str, reference: str, **ignored_kwargs: Any
    ) -> score_result.ScoreResult:
        """
        Calculate the score based on whether the output string exactly matches the expected output.

        Args:
            output: The output string to check.
            reference: The expected output string to compare against.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with a value of 1.0 if the strings match,
                0.0 otherwise.
        """
        value_left = output if self._case_sensitive else output.lower()
        value_right = reference if self._case_sensitive else reference.lower()

        if value_left == value_right:
            return score_result.ScoreResult(value=1.0, name=self.name)

        return score_result.ScoreResult(value=0.0, name=self.name)



from .api import (
    enabled_in_config,
    randomized_should_enable_reporting,
    setup_sentry_error_tracker,
)

__all__ = [
    "enabled_in_config",
    "randomized_should_enable_reporting",
    "setup_sentry_error_tracker",
]



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.experiment_item import ExperimentItem
from ..types.experiment_item_public import ExperimentItemPublic
from ..types.experiment_page_public import ExperimentPagePublic
from ..types.experiment_public import ExperimentPublic
from ..types.json_node_write import JsonNodeWrite
from ..types.prompt_version_link_write import PromptVersionLinkWrite
from .raw_client import AsyncRawExperimentsClient, RawExperimentsClient
from .types.experiment_write_type import ExperimentWriteType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ExperimentsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawExperimentsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawExperimentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawExperimentsClient
        """
        return self._raw_client

    def find_experiments(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        optimization_id: typing.Optional[str] = None,
        types: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        prompt_id: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ExperimentPagePublic:
        """
        Find experiments

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        optimization_id : typing.Optional[str]

        types : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        prompt_id : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentPagePublic
            Experiments resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.find_experiments()
        """
        _response = self._raw_client.find_experiments(
            page=page,
            size=size,
            dataset_id=dataset_id,
            optimization_id=optimization_id,
            types=types,
            name=name,
            dataset_deleted=dataset_deleted,
            prompt_id=prompt_id,
            sorting=sorting,
            request_options=request_options,
        )
        return _response.data

    def create_experiment(
        self,
        *,
        dataset_name: str,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        type: typing.Optional[ExperimentWriteType] = OMIT,
        optimization_id: typing.Optional[str] = OMIT,
        prompt_version: typing.Optional[PromptVersionLinkWrite] = OMIT,
        prompt_versions: typing.Optional[typing.Sequence[PromptVersionLinkWrite]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create experiment

        Parameters
        ----------
        dataset_name : str

        id : typing.Optional[str]

        name : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        type : typing.Optional[ExperimentWriteType]

        optimization_id : typing.Optional[str]

        prompt_version : typing.Optional[PromptVersionLinkWrite]

        prompt_versions : typing.Optional[typing.Sequence[PromptVersionLinkWrite]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.create_experiment(dataset_name='dataset_name', )
        """
        _response = self._raw_client.create_experiment(
            dataset_name=dataset_name,
            id=id,
            name=name,
            metadata=metadata,
            type=type,
            optimization_id=optimization_id,
            prompt_version=prompt_version,
            prompt_versions=prompt_versions,
            request_options=request_options,
        )
        return _response.data

    def create_experiment_items(
        self,
        *,
        experiment_items: typing.Sequence[ExperimentItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create experiment items

        Parameters
        ----------
        experiment_items : typing.Sequence[ExperimentItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import ExperimentItem
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.create_experiment_items(experiment_items=[ExperimentItem(experiment_id='experiment_id', dataset_item_id='dataset_item_id', trace_id='trace_id', )], )
        """
        _response = self._raw_client.create_experiment_items(
            experiment_items=experiment_items, request_options=request_options
        )
        return _response.data

    def delete_experiment_items(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete experiment items

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.delete_experiment_items(ids=['ids'], )
        """
        _response = self._raw_client.delete_experiment_items(ids=ids, request_options=request_options)
        return _response.data

    def delete_experiments_by_id(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete experiments by id

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.delete_experiments_by_id(ids=['ids'], )
        """
        _response = self._raw_client.delete_experiments_by_id(ids=ids, request_options=request_options)
        return _response.data

    def find_feedback_score_names(
        self, *, experiment_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[str]:
        """
        Find Feedback Score names

        Parameters
        ----------
        experiment_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[str]
            Feedback Scores resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.find_feedback_score_names()
        """
        _response = self._raw_client.find_feedback_score_names(
            experiment_ids=experiment_ids, request_options=request_options
        )
        return _response.data

    def get_experiment_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ExperimentPublic:
        """
        Get experiment by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentPublic
            Experiment resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.get_experiment_by_id(id='id', )
        """
        _response = self._raw_client.get_experiment_by_id(id, request_options=request_options)
        return _response.data

    def get_experiment_item_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ExperimentItemPublic:
        """
        Get experiment item by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentItemPublic
            Experiment item resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.experiments.get_experiment_item_by_id(id='id', )
        """
        _response = self._raw_client.get_experiment_item_by_id(id, request_options=request_options)
        return _response.data

    def stream_experiment_items(
        self,
        *,
        experiment_name: str,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        truncate: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[bytes]:
        """
        Stream experiment items

        Parameters
        ----------
        experiment_name : str

        limit : typing.Optional[int]

        last_retrieved_id : typing.Optional[str]

        truncate : typing.Optional[bool]
            Truncate image included in either input, output or metadata

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[bytes]
            Experiment items stream or error during process
        """
        with self._raw_client.stream_experiment_items(
            experiment_name=experiment_name,
            limit=limit,
            last_retrieved_id=last_retrieved_id,
            truncate=truncate,
            request_options=request_options,
        ) as r:
            yield from r.data

    def stream_experiments(
        self,
        *,
        name: str,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[bytes]:
        """
        Stream experiments

        Parameters
        ----------
        name : str

        limit : typing.Optional[int]

        last_retrieved_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[bytes]
            Experiments stream or error during process
        """
        with self._raw_client.stream_experiments(
            name=name, limit=limit, last_retrieved_id=last_retrieved_id, request_options=request_options
        ) as r:
            yield from r.data


class AsyncExperimentsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawExperimentsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawExperimentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawExperimentsClient
        """
        return self._raw_client

    async def find_experiments(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        dataset_id: typing.Optional[str] = None,
        optimization_id: typing.Optional[str] = None,
        types: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        dataset_deleted: typing.Optional[bool] = None,
        prompt_id: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ExperimentPagePublic:
        """
        Find experiments

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        dataset_id : typing.Optional[str]

        optimization_id : typing.Optional[str]

        types : typing.Optional[str]

        name : typing.Optional[str]

        dataset_deleted : typing.Optional[bool]

        prompt_id : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentPagePublic
            Experiments resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.find_experiments()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_experiments(
            page=page,
            size=size,
            dataset_id=dataset_id,
            optimization_id=optimization_id,
            types=types,
            name=name,
            dataset_deleted=dataset_deleted,
            prompt_id=prompt_id,
            sorting=sorting,
            request_options=request_options,
        )
        return _response.data

    async def create_experiment(
        self,
        *,
        dataset_name: str,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        type: typing.Optional[ExperimentWriteType] = OMIT,
        optimization_id: typing.Optional[str] = OMIT,
        prompt_version: typing.Optional[PromptVersionLinkWrite] = OMIT,
        prompt_versions: typing.Optional[typing.Sequence[PromptVersionLinkWrite]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create experiment

        Parameters
        ----------
        dataset_name : str

        id : typing.Optional[str]

        name : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        type : typing.Optional[ExperimentWriteType]

        optimization_id : typing.Optional[str]

        prompt_version : typing.Optional[PromptVersionLinkWrite]

        prompt_versions : typing.Optional[typing.Sequence[PromptVersionLinkWrite]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.create_experiment(dataset_name='dataset_name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_experiment(
            dataset_name=dataset_name,
            id=id,
            name=name,
            metadata=metadata,
            type=type,
            optimization_id=optimization_id,
            prompt_version=prompt_version,
            prompt_versions=prompt_versions,
            request_options=request_options,
        )
        return _response.data

    async def create_experiment_items(
        self,
        *,
        experiment_items: typing.Sequence[ExperimentItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create experiment items

        Parameters
        ----------
        experiment_items : typing.Sequence[ExperimentItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import ExperimentItem
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.create_experiment_items(experiment_items=[ExperimentItem(experiment_id='experiment_id', dataset_item_id='dataset_item_id', trace_id='trace_id', )], )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_experiment_items(
            experiment_items=experiment_items, request_options=request_options
        )
        return _response.data

    async def delete_experiment_items(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete experiment items

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.delete_experiment_items(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_experiment_items(ids=ids, request_options=request_options)
        return _response.data

    async def delete_experiments_by_id(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete experiments by id

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.delete_experiments_by_id(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_experiments_by_id(ids=ids, request_options=request_options)
        return _response.data

    async def find_feedback_score_names(
        self, *, experiment_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[str]:
        """
        Find Feedback Score names

        Parameters
        ----------
        experiment_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[str]
            Feedback Scores resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.find_feedback_score_names()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_feedback_score_names(
            experiment_ids=experiment_ids, request_options=request_options
        )
        return _response.data

    async def get_experiment_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ExperimentPublic:
        """
        Get experiment by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentPublic
            Experiment resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.get_experiment_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_experiment_by_id(id, request_options=request_options)
        return _response.data

    async def get_experiment_item_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> ExperimentItemPublic:
        """
        Get experiment item by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ExperimentItemPublic
            Experiment item resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.experiments.get_experiment_item_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_experiment_item_by_id(id, request_options=request_options)
        return _response.data

    async def stream_experiment_items(
        self,
        *,
        experiment_name: str,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        truncate: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[bytes]:
        """
        Stream experiment items

        Parameters
        ----------
        experiment_name : str

        limit : typing.Optional[int]

        last_retrieved_id : typing.Optional[str]

        truncate : typing.Optional[bool]
            Truncate image included in either input, output or metadata

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[bytes]
            Experiment items stream or error during process
        """
        async with self._raw_client.stream_experiment_items(
            experiment_name=experiment_name,
            limit=limit,
            last_retrieved_id=last_retrieved_id,
            truncate=truncate,
            request_options=request_options,
        ) as r:
            async for data in r.data:
                yield data

    async def stream_experiments(
        self,
        *,
        name: str,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[bytes]:
        """
        Stream experiments

        Parameters
        ----------
        name : str

        limit : typing.Optional[int]

        last_retrieved_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[bytes]
            Experiments stream or error during process
        """
        async with self._raw_client.stream_experiments(
            name=name, limit=limit, last_retrieved_id=last_retrieved_id, request_options=request_options
        ) as r:
            async for data in r.data:
                yield data



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .assistant_message import AssistantMessage
from .assistant_message_role import AssistantMessageRole
from .attachment import Attachment
from .attachment_page import AttachmentPage
from .auth_details_holder import AuthDetailsHolder
from .automation_rule_evaluator import (
    AutomationRuleEvaluator,
    AutomationRuleEvaluator_LlmAsJudge,
    AutomationRuleEvaluator_UserDefinedMetricPython,
)
from .automation_rule_evaluator_llm_as_judge import AutomationRuleEvaluatorLlmAsJudge
from .automation_rule_evaluator_llm_as_judge_public import AutomationRuleEvaluatorLlmAsJudgePublic
from .automation_rule_evaluator_llm_as_judge_write import AutomationRuleEvaluatorLlmAsJudgeWrite
from .automation_rule_evaluator_object_public import (
    AutomationRuleEvaluatorObjectPublic,
    AutomationRuleEvaluatorObjectPublic_LlmAsJudge,
    AutomationRuleEvaluatorObjectPublic_UserDefinedMetricPython,
)
from .automation_rule_evaluator_page_public import AutomationRuleEvaluatorPagePublic
from .automation_rule_evaluator_public import (
    AutomationRuleEvaluatorPublic,
    AutomationRuleEvaluatorPublic_LlmAsJudge,
    AutomationRuleEvaluatorPublic_UserDefinedMetricPython,
)
from .automation_rule_evaluator_update import (
    AutomationRuleEvaluatorUpdate,
    AutomationRuleEvaluatorUpdate_LlmAsJudge,
    AutomationRuleEvaluatorUpdate_UserDefinedMetricPython,
)
from .automation_rule_evaluator_update_llm_as_judge import AutomationRuleEvaluatorUpdateLlmAsJudge
from .automation_rule_evaluator_update_user_defined_metric_python import (
    AutomationRuleEvaluatorUpdateUserDefinedMetricPython,
)
from .automation_rule_evaluator_user_defined_metric_python import AutomationRuleEvaluatorUserDefinedMetricPython
from .automation_rule_evaluator_user_defined_metric_python_public import (
    AutomationRuleEvaluatorUserDefinedMetricPythonPublic,
)
from .automation_rule_evaluator_user_defined_metric_python_write import (
    AutomationRuleEvaluatorUserDefinedMetricPythonWrite,
)
from .automation_rule_evaluator_write import (
    AutomationRuleEvaluatorWrite,
    AutomationRuleEvaluatorWrite_LlmAsJudge,
    AutomationRuleEvaluatorWrite_UserDefinedMetricPython,
)
from .avg_value_stat_public import AvgValueStatPublic
from .batch_delete import BatchDelete
from .bi_information import BiInformation
from .bi_information_response import BiInformationResponse
from .categorical_feedback_definition import CategoricalFeedbackDefinition
from .categorical_feedback_definition_create import CategoricalFeedbackDefinitionCreate
from .categorical_feedback_definition_public import CategoricalFeedbackDefinitionPublic
from .categorical_feedback_definition_update import CategoricalFeedbackDefinitionUpdate
from .categorical_feedback_detail import CategoricalFeedbackDetail
from .categorical_feedback_detail_create import CategoricalFeedbackDetailCreate
from .categorical_feedback_detail_public import CategoricalFeedbackDetailPublic
from .categorical_feedback_detail_update import CategoricalFeedbackDetailUpdate
from .chat_completion_choice import ChatCompletionChoice
from .chat_completion_response import ChatCompletionResponse
from .check import Check
from .check_name import CheckName
from .check_public import CheckPublic
from .check_public_name import CheckPublicName
from .check_public_result import CheckPublicResult
from .check_result import CheckResult
from .chunked_output_json_node import ChunkedOutputJsonNode
from .chunked_output_json_node_public import ChunkedOutputJsonNodePublic
from .chunked_output_json_node_public_type import ChunkedOutputJsonNodePublicType
from .chunked_output_json_node_type import ChunkedOutputJsonNodeType
from .column import Column
from .column_compare import ColumnCompare
from .column_compare_types_item import ColumnCompareTypesItem
from .column_public import ColumnPublic
from .column_public_types_item import ColumnPublicTypesItem
from .column_types_item import ColumnTypesItem
from .comment import Comment
from .comment_compare import CommentCompare
from .comment_public import CommentPublic
from .complete_multipart_upload_request import CompleteMultipartUploadRequest
from .complete_multipart_upload_request_entity_type import CompleteMultipartUploadRequestEntityType
from .completion_tokens_details import CompletionTokensDetails
from .count_value_stat_public import CountValueStatPublic
from .data_point_number_public import DataPointNumberPublic
from .dataset import Dataset
from .dataset_item import DatasetItem
from .dataset_item_batch import DatasetItemBatch
from .dataset_item_compare import DatasetItemCompare
from .dataset_item_compare_source import DatasetItemCompareSource
from .dataset_item_page_compare import DatasetItemPageCompare
from .dataset_item_page_public import DatasetItemPagePublic
from .dataset_item_public import DatasetItemPublic
from .dataset_item_public_source import DatasetItemPublicSource
from .dataset_item_source import DatasetItemSource
from .dataset_item_write import DatasetItemWrite
from .dataset_item_write_source import DatasetItemWriteSource
from .dataset_page_public import DatasetPagePublic
from .dataset_public import DatasetPublic
from .dataset_public_visibility import DatasetPublicVisibility
from .dataset_visibility import DatasetVisibility
from .delete_attachments_request import DeleteAttachmentsRequest
from .delete_attachments_request_entity_type import DeleteAttachmentsRequestEntityType
from .delete_feedback_score import DeleteFeedbackScore
from .delete_ids_holder import DeleteIdsHolder
from .delta import Delta
from .error_info import ErrorInfo
from .error_info_public import ErrorInfoPublic
from .error_info_write import ErrorInfoWrite
from .error_message import ErrorMessage
from .error_message_detail import ErrorMessageDetail
from .error_message_detailed import ErrorMessageDetailed
from .error_message_public import ErrorMessagePublic
from .experiment import Experiment
from .experiment_item import ExperimentItem
from .experiment_item_compare import ExperimentItemCompare
from .experiment_item_public import ExperimentItemPublic
from .experiment_page_public import ExperimentPagePublic
from .experiment_public import ExperimentPublic
from .experiment_public_type import ExperimentPublicType
from .experiment_type import ExperimentType
from .export_trace_service_request import ExportTraceServiceRequest
from .feedback import Feedback, Feedback_Categorical, Feedback_Numerical
from .feedback_create import FeedbackCreate, FeedbackCreate_Categorical, FeedbackCreate_Numerical
from .feedback_definition_page_public import FeedbackDefinitionPagePublic
from .feedback_object_public import (
    FeedbackObjectPublic,
    FeedbackObjectPublic_Categorical,
    FeedbackObjectPublic_Numerical,
)
from .feedback_public import FeedbackPublic, FeedbackPublic_Categorical, FeedbackPublic_Numerical
from .feedback_score import FeedbackScore
from .feedback_score_average import FeedbackScoreAverage
from .feedback_score_average_detailed import FeedbackScoreAverageDetailed
from .feedback_score_average_public import FeedbackScoreAveragePublic
from .feedback_score_batch import FeedbackScoreBatch
from .feedback_score_batch_item import FeedbackScoreBatchItem
from .feedback_score_batch_item_source import FeedbackScoreBatchItemSource
from .feedback_score_compare import FeedbackScoreCompare
from .feedback_score_compare_source import FeedbackScoreCompareSource
from .feedback_score_names import FeedbackScoreNames
from .feedback_score_public import FeedbackScorePublic
from .feedback_score_public_source import FeedbackScorePublicSource
from .feedback_score_source import FeedbackScoreSource
from .feedback_update import FeedbackUpdate, FeedbackUpdate_Categorical, FeedbackUpdate_Numerical
from .function import Function
from .function_call import FunctionCall
from .guardrail import Guardrail
from .guardrail_batch import GuardrailBatch
from .guardrail_name import GuardrailName
from .guardrail_result import GuardrailResult
from .guardrail_write import GuardrailWrite
from .guardrail_write_name import GuardrailWriteName
from .guardrail_write_result import GuardrailWriteResult
from .guardrails_validation import GuardrailsValidation
from .guardrails_validation_public import GuardrailsValidationPublic
from .json_list_string import JsonListString
from .json_list_string_compare import JsonListStringCompare
from .json_list_string_public import JsonListStringPublic
from .json_list_string_write import JsonListStringWrite
from .json_node import JsonNode
from .json_node_detail import JsonNodeDetail
from .json_node_public import JsonNodePublic
from .json_node_write import JsonNodeWrite
from .json_schema import JsonSchema
from .llm_as_judge_code import LlmAsJudgeCode
from .llm_as_judge_code_public import LlmAsJudgeCodePublic
from .llm_as_judge_code_write import LlmAsJudgeCodeWrite
from .llm_as_judge_message import LlmAsJudgeMessage
from .llm_as_judge_message_public import LlmAsJudgeMessagePublic
from .llm_as_judge_message_public_role import LlmAsJudgeMessagePublicRole
from .llm_as_judge_message_role import LlmAsJudgeMessageRole
from .llm_as_judge_message_write import LlmAsJudgeMessageWrite
from .llm_as_judge_message_write_role import LlmAsJudgeMessageWriteRole
from .llm_as_judge_model_parameters import LlmAsJudgeModelParameters
from .llm_as_judge_model_parameters_public import LlmAsJudgeModelParametersPublic
from .llm_as_judge_model_parameters_write import LlmAsJudgeModelParametersWrite
from .llm_as_judge_output_schema import LlmAsJudgeOutputSchema
from .llm_as_judge_output_schema_public import LlmAsJudgeOutputSchemaPublic
from .llm_as_judge_output_schema_public_type import LlmAsJudgeOutputSchemaPublicType
from .llm_as_judge_output_schema_type import LlmAsJudgeOutputSchemaType
from .llm_as_judge_output_schema_write import LlmAsJudgeOutputSchemaWrite
from .llm_as_judge_output_schema_write_type import LlmAsJudgeOutputSchemaWriteType
from .log_item import LogItem
from .log_item_level import LogItemLevel
from .log_page import LogPage
from .message import Message
from .multipart_upload_part import MultipartUploadPart
from .numerical_feedback_definition import NumericalFeedbackDefinition
from .numerical_feedback_definition_create import NumericalFeedbackDefinitionCreate
from .numerical_feedback_definition_public import NumericalFeedbackDefinitionPublic
from .numerical_feedback_definition_update import NumericalFeedbackDefinitionUpdate
from .numerical_feedback_detail import NumericalFeedbackDetail
from .numerical_feedback_detail_create import NumericalFeedbackDetailCreate
from .numerical_feedback_detail_public import NumericalFeedbackDetailPublic
from .numerical_feedback_detail_update import NumericalFeedbackDetailUpdate
from .optimization import Optimization
from .optimization_page_public import OptimizationPagePublic
from .optimization_public import OptimizationPublic
from .optimization_public_status import OptimizationPublicStatus
from .optimization_status import OptimizationStatus
from .page_columns import PageColumns
from .percentage_value_stat_public import PercentageValueStatPublic
from .percentage_values import PercentageValues
from .percentage_values_detailed import PercentageValuesDetailed
from .percentage_values_public import PercentageValuesPublic
from .project import Project
from .project_detailed import ProjectDetailed
from .project_detailed_visibility import ProjectDetailedVisibility
from .project_metric_response_public import ProjectMetricResponsePublic
from .project_metric_response_public_interval import ProjectMetricResponsePublicInterval
from .project_metric_response_public_metric_type import ProjectMetricResponsePublicMetricType
from .project_page_public import ProjectPagePublic
from .project_public import ProjectPublic
from .project_public_visibility import ProjectPublicVisibility
from .project_stat_item_object_public import (
    ProjectStatItemObjectPublic,
    ProjectStatItemObjectPublic_Avg,
    ProjectStatItemObjectPublic_Count,
    ProjectStatItemObjectPublic_Percentage,
)
from .project_stats_public import ProjectStatsPublic
from .project_stats_summary import ProjectStatsSummary
from .project_stats_summary_item import ProjectStatsSummaryItem
from .project_visibility import ProjectVisibility
from .prompt import Prompt
from .prompt_detail import PromptDetail
from .prompt_page_public import PromptPagePublic
from .prompt_public import PromptPublic
from .prompt_tokens_details import PromptTokensDetails
from .prompt_type import PromptType
from .prompt_version import PromptVersion
from .prompt_version_detail import PromptVersionDetail
from .prompt_version_detail_type import PromptVersionDetailType
from .prompt_version_link import PromptVersionLink
from .prompt_version_link_public import PromptVersionLinkPublic
from .prompt_version_link_write import PromptVersionLinkWrite
from .prompt_version_page_public import PromptVersionPagePublic
from .prompt_version_public import PromptVersionPublic
from .prompt_version_public_type import PromptVersionPublicType
from .prompt_version_type import PromptVersionType
from .provider_api_key import ProviderApiKey
from .provider_api_key_page_public import ProviderApiKeyPagePublic
from .provider_api_key_provider import ProviderApiKeyProvider
from .provider_api_key_public import ProviderApiKeyPublic
from .provider_api_key_public_provider import ProviderApiKeyPublicProvider
from .response_format import ResponseFormat
from .response_format_type import ResponseFormatType
from .results_number_public import ResultsNumberPublic
from .score_name import ScoreName
from .service_toggles_config import ServiceTogglesConfig
from .span import Span
from .span_batch import SpanBatch
from .span_filter_public import SpanFilterPublic
from .span_filter_public_operator import SpanFilterPublicOperator
from .span_page_public import SpanPagePublic
from .span_public import SpanPublic
from .span_public_type import SpanPublicType
from .span_type import SpanType
from .span_write import SpanWrite
from .span_write_type import SpanWriteType
from .spans_count_response import SpansCountResponse
from .start_multipart_upload_response import StartMultipartUploadResponse
from .stream_options import StreamOptions
from .tool import Tool
from .tool_call import ToolCall
from .trace import Trace
from .trace_batch import TraceBatch
from .trace_count_response import TraceCountResponse
from .trace_filter_public import TraceFilterPublic
from .trace_filter_public_operator import TraceFilterPublicOperator
from .trace_page_public import TracePagePublic
from .trace_public import TracePublic
from .trace_thread import TraceThread
from .trace_thread_page import TraceThreadPage
from .trace_write import TraceWrite
from .usage import Usage
from .user_defined_metric_python_code import UserDefinedMetricPythonCode
from .user_defined_metric_python_code_public import UserDefinedMetricPythonCodePublic
from .user_defined_metric_python_code_write import UserDefinedMetricPythonCodeWrite
from .workspace_name_holder import WorkspaceNameHolder
from .workspace_spans_count import WorkspaceSpansCount
from .workspace_trace_count import WorkspaceTraceCount

__all__ = [
    "AssistantMessage",
    "AssistantMessageRole",
    "Attachment",
    "AttachmentPage",
    "AuthDetailsHolder",
    "AutomationRuleEvaluator",
    "AutomationRuleEvaluatorLlmAsJudge",
    "AutomationRuleEvaluatorLlmAsJudgePublic",
    "AutomationRuleEvaluatorLlmAsJudgeWrite",
    "AutomationRuleEvaluatorObjectPublic",
    "AutomationRuleEvaluatorObjectPublic_LlmAsJudge",
    "AutomationRuleEvaluatorObjectPublic_UserDefinedMetricPython",
    "AutomationRuleEvaluatorPagePublic",
    "AutomationRuleEvaluatorPublic",
    "AutomationRuleEvaluatorPublic_LlmAsJudge",
    "AutomationRuleEvaluatorPublic_UserDefinedMetricPython",
    "AutomationRuleEvaluatorUpdate",
    "AutomationRuleEvaluatorUpdateLlmAsJudge",
    "AutomationRuleEvaluatorUpdateUserDefinedMetricPython",
    "AutomationRuleEvaluatorUpdate_LlmAsJudge",
    "AutomationRuleEvaluatorUpdate_UserDefinedMetricPython",
    "AutomationRuleEvaluatorUserDefinedMetricPython",
    "AutomationRuleEvaluatorUserDefinedMetricPythonPublic",
    "AutomationRuleEvaluatorUserDefinedMetricPythonWrite",
    "AutomationRuleEvaluatorWrite",
    "AutomationRuleEvaluatorWrite_LlmAsJudge",
    "AutomationRuleEvaluatorWrite_UserDefinedMetricPython",
    "AutomationRuleEvaluator_LlmAsJudge",
    "AutomationRuleEvaluator_UserDefinedMetricPython",
    "AvgValueStatPublic",
    "BatchDelete",
    "BiInformation",
    "BiInformationResponse",
    "CategoricalFeedbackDefinition",
    "CategoricalFeedbackDefinitionCreate",
    "CategoricalFeedbackDefinitionPublic",
    "CategoricalFeedbackDefinitionUpdate",
    "CategoricalFeedbackDetail",
    "CategoricalFeedbackDetailCreate",
    "CategoricalFeedbackDetailPublic",
    "CategoricalFeedbackDetailUpdate",
    "ChatCompletionChoice",
    "ChatCompletionResponse",
    "Check",
    "CheckName",
    "CheckPublic",
    "CheckPublicName",
    "CheckPublicResult",
    "CheckResult",
    "ChunkedOutputJsonNode",
    "ChunkedOutputJsonNodePublic",
    "ChunkedOutputJsonNodePublicType",
    "ChunkedOutputJsonNodeType",
    "Column",
    "ColumnCompare",
    "ColumnCompareTypesItem",
    "ColumnPublic",
    "ColumnPublicTypesItem",
    "ColumnTypesItem",
    "Comment",
    "CommentCompare",
    "CommentPublic",
    "CompleteMultipartUploadRequest",
    "CompleteMultipartUploadRequestEntityType",
    "CompletionTokensDetails",
    "CountValueStatPublic",
    "DataPointNumberPublic",
    "Dataset",
    "DatasetItem",
    "DatasetItemBatch",
    "DatasetItemCompare",
    "DatasetItemCompareSource",
    "DatasetItemPageCompare",
    "DatasetItemPagePublic",
    "DatasetItemPublic",
    "DatasetItemPublicSource",
    "DatasetItemSource",
    "DatasetItemWrite",
    "DatasetItemWriteSource",
    "DatasetPagePublic",
    "DatasetPublic",
    "DatasetPublicVisibility",
    "DatasetVisibility",
    "DeleteAttachmentsRequest",
    "DeleteAttachmentsRequestEntityType",
    "DeleteFeedbackScore",
    "DeleteIdsHolder",
    "Delta",
    "ErrorInfo",
    "ErrorInfoPublic",
    "ErrorInfoWrite",
    "ErrorMessage",
    "ErrorMessageDetail",
    "ErrorMessageDetailed",
    "ErrorMessagePublic",
    "Experiment",
    "ExperimentItem",
    "ExperimentItemCompare",
    "ExperimentItemPublic",
    "ExperimentPagePublic",
    "ExperimentPublic",
    "ExperimentPublicType",
    "ExperimentType",
    "ExportTraceServiceRequest",
    "Feedback",
    "FeedbackCreate",
    "FeedbackCreate_Categorical",
    "FeedbackCreate_Numerical",
    "FeedbackDefinitionPagePublic",
    "FeedbackObjectPublic",
    "FeedbackObjectPublic_Categorical",
    "FeedbackObjectPublic_Numerical",
    "FeedbackPublic",
    "FeedbackPublic_Categorical",
    "FeedbackPublic_Numerical",
    "FeedbackScore",
    "FeedbackScoreAverage",
    "FeedbackScoreAverageDetailed",
    "FeedbackScoreAveragePublic",
    "FeedbackScoreBatch",
    "FeedbackScoreBatchItem",
    "FeedbackScoreBatchItemSource",
    "FeedbackScoreCompare",
    "FeedbackScoreCompareSource",
    "FeedbackScoreNames",
    "FeedbackScorePublic",
    "FeedbackScorePublicSource",
    "FeedbackScoreSource",
    "FeedbackUpdate",
    "FeedbackUpdate_Categorical",
    "FeedbackUpdate_Numerical",
    "Feedback_Categorical",
    "Feedback_Numerical",
    "Function",
    "FunctionCall",
    "Guardrail",
    "GuardrailBatch",
    "GuardrailName",
    "GuardrailResult",
    "GuardrailWrite",
    "GuardrailWriteName",
    "GuardrailWriteResult",
    "GuardrailsValidation",
    "GuardrailsValidationPublic",
    "JsonListString",
    "JsonListStringCompare",
    "JsonListStringPublic",
    "JsonListStringWrite",
    "JsonNode",
    "JsonNodeDetail",
    "JsonNodePublic",
    "JsonNodeWrite",
    "JsonSchema",
    "LlmAsJudgeCode",
    "LlmAsJudgeCodePublic",
    "LlmAsJudgeCodeWrite",
    "LlmAsJudgeMessage",
    "LlmAsJudgeMessagePublic",
    "LlmAsJudgeMessagePublicRole",
    "LlmAsJudgeMessageRole",
    "LlmAsJudgeMessageWrite",
    "LlmAsJudgeMessageWriteRole",
    "LlmAsJudgeModelParameters",
    "LlmAsJudgeModelParametersPublic",
    "LlmAsJudgeModelParametersWrite",
    "LlmAsJudgeOutputSchema",
    "LlmAsJudgeOutputSchemaPublic",
    "LlmAsJudgeOutputSchemaPublicType",
    "LlmAsJudgeOutputSchemaType",
    "LlmAsJudgeOutputSchemaWrite",
    "LlmAsJudgeOutputSchemaWriteType",
    "LogItem",
    "LogItemLevel",
    "LogPage",
    "Message",
    "MultipartUploadPart",
    "NumericalFeedbackDefinition",
    "NumericalFeedbackDefinitionCreate",
    "NumericalFeedbackDefinitionPublic",
    "NumericalFeedbackDefinitionUpdate",
    "NumericalFeedbackDetail",
    "NumericalFeedbackDetailCreate",
    "NumericalFeedbackDetailPublic",
    "NumericalFeedbackDetailUpdate",
    "Optimization",
    "OptimizationPagePublic",
    "OptimizationPublic",
    "OptimizationPublicStatus",
    "OptimizationStatus",
    "PageColumns",
    "PercentageValueStatPublic",
    "PercentageValues",
    "PercentageValuesDetailed",
    "PercentageValuesPublic",
    "Project",
    "ProjectDetailed",
    "ProjectDetailedVisibility",
    "ProjectMetricResponsePublic",
    "ProjectMetricResponsePublicInterval",
    "ProjectMetricResponsePublicMetricType",
    "ProjectPagePublic",
    "ProjectPublic",
    "ProjectPublicVisibility",
    "ProjectStatItemObjectPublic",
    "ProjectStatItemObjectPublic_Avg",
    "ProjectStatItemObjectPublic_Count",
    "ProjectStatItemObjectPublic_Percentage",
    "ProjectStatsPublic",
    "ProjectStatsSummary",
    "ProjectStatsSummaryItem",
    "ProjectVisibility",
    "Prompt",
    "PromptDetail",
    "PromptPagePublic",
    "PromptPublic",
    "PromptTokensDetails",
    "PromptType",
    "PromptVersion",
    "PromptVersionDetail",
    "PromptVersionDetailType",
    "PromptVersionLink",
    "PromptVersionLinkPublic",
    "PromptVersionLinkWrite",
    "PromptVersionPagePublic",
    "PromptVersionPublic",
    "PromptVersionPublicType",
    "PromptVersionType",
    "ProviderApiKey",
    "ProviderApiKeyPagePublic",
    "ProviderApiKeyProvider",
    "ProviderApiKeyPublic",
    "ProviderApiKeyPublicProvider",
    "ResponseFormat",
    "ResponseFormatType",
    "ResultsNumberPublic",
    "ScoreName",
    "ServiceTogglesConfig",
    "Span",
    "SpanBatch",
    "SpanFilterPublic",
    "SpanFilterPublicOperator",
    "SpanPagePublic",
    "SpanPublic",
    "SpanPublicType",
    "SpanType",
    "SpanWrite",
    "SpanWriteType",
    "SpansCountResponse",
    "StartMultipartUploadResponse",
    "StreamOptions",
    "Tool",
    "ToolCall",
    "Trace",
    "TraceBatch",
    "TraceCountResponse",
    "TraceFilterPublic",
    "TraceFilterPublicOperator",
    "TracePagePublic",
    "TracePublic",
    "TraceThread",
    "TraceThreadPage",
    "TraceWrite",
    "Usage",
    "UserDefinedMetricPythonCode",
    "UserDefinedMetricPythonCodePublic",
    "UserDefinedMetricPythonCodeWrite",
    "WorkspaceNameHolder",
    "WorkspaceSpansCount",
    "WorkspaceTraceCount",
]



from .. import rest_api
from . import public_methods_patcher, retry_decorator


def configure(rest_client: rest_api.OpikApi) -> None:
    _configure_retries(rest_client)


def _configure_retries(rest_client: rest_api.OpikApi) -> None:
    domain_client_names = [
        "attachments",
        "datasets",
        "experiments",
        "traces",
        "spans",
        "projects",
        "prompts",
    ]
    for domain_client_name in domain_client_names:
        public_methods_patcher.patch(
            getattr(rest_client, domain_client_name), retry_decorator.opik_rest_retry
        )



import logging
from typing import Iterator, AsyncIterator, List, Optional, Callable, TypeVar

from opik.types import ErrorInfoDict
from opik.api_objects import trace, span
from opik.decorator import generator_wrappers, error_info_collector
import functools
import openai

LOGGER = logging.getLogger(__name__)

original_stream_iter_method = openai.Stream.__iter__
original_async_stream_aiter_method = openai.AsyncStream.__aiter__

StreamItem = TypeVar("StreamItem")
AggregatedResult = TypeVar("AggregatedResult")


def patch_sync_stream(
    stream: openai.Stream,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    generations_aggregator: Callable[[List[StreamItem]], Optional[AggregatedResult]],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> openai.Stream:
    """
    Used in the following cases
    ```
    stream = client.messages.create(stream=True)
    for event in stream:
        print(event)
    ```

    """

    def Stream__iter__decorator(dunder_iter_func: Callable) -> Callable:
        @functools.wraps(dunder_iter_func)
        def wrapper(
            self: openai.Stream,
        ) -> Iterator[StreamItem]:
            try:
                accumulated_items: List[StreamItem] = []
                error_info: Optional[ErrorInfoDict] = None
                for item in dunder_iter_func(self):
                    accumulated_items.append(item)
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from openai.Stream.",
                    str(exception),
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")
                output = (
                    generations_aggregator(accumulated_items)
                    if error_info is None
                    else None
                )
                finally_callback(
                    output=output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    openai.Stream.__iter__ = Stream__iter__decorator(original_stream_iter_method)

    stream.opik_tracked_instance = True
    stream.span_to_end = span_to_end
    stream.trace_to_end = trace_to_end

    return stream


def patch_async_stream(
    stream: openai.AsyncStream,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    generations_aggregator: Callable[[List[StreamItem]], Optional[AggregatedResult]],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> openai.Stream:
    """
    Used in the following cases
    ```
    astream = async_client.messages.create(stream=True)
    async for event in astream:
        print(event)
    ```
    """

    def AsyncStream__aiter__decorator(dunder_aiter_func: Callable) -> Callable:
        @functools.wraps(dunder_aiter_func)
        async def wrapper(
            self: openai.AsyncStream,
        ) -> AsyncIterator[StreamItem]:
            try:
                accumulated_items: List[StreamItem] = []
                error_info: Optional[ErrorInfoDict] = None

                async for item in dunder_aiter_func(self):
                    accumulated_items.append(item)
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from openai.AsyncStream.",
                    str(exception),
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")
                output = (
                    generations_aggregator(accumulated_items)
                    if error_info is None
                    else None
                )
                finally_callback(
                    output=output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    openai.AsyncStream.__aiter__ = AsyncStream__aiter__decorator(
        original_async_stream_aiter_method
    )

    stream.opik_tracked_instance = True
    stream.span_to_end = span_to_end
    stream.trace_to_end = trace_to_end

    return stream



# This file was auto-generated by Fern from our API Definition.

import typing

ProjectMetricResponsePublicMetricType = typing.Union[
    typing.Literal["FEEDBACK_SCORES", "TRACE_COUNT", "TOKEN_USAGE", "DURATION", "COST", "GUARDRAILS_FAILED_COUNT"],
    typing.Any,
]



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .function_call import FunctionCall
from .tool_call import ToolCall


class Delta(UniversalBaseModel):
    role: typing.Optional[str] = None
    content: typing.Optional[str] = None
    tool_calls: typing.Optional[typing.List[ToolCall]] = None
    function_call: typing.Optional[FunctionCall] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
